// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"crypto/x509"
	"encoding/pem"
	"fmt"
	"io/ioutil"
	"os"

	"github.com/spf13/cobra"

	xbus "bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
)

// registerCmd represents the "register" command.
var registerCmd = &cobra.Command{
	Use:   "register",
	Short: "Generate a CSR and ask Xbus for a registration.",
	Long: `Generate a Certificate Signing Request then send it over to Xbus through the
anonymouse "registration" channel.`,
	Run: register,
}

// register parses arguments to proceed with the registration.
func register(cmd *cobra.Command, args []string) {
	os.Exit(doRegister(false, false))
}

// loadPrivKey loads the private key stored in the specified file.
func loadPrivKey(privKeyPath string) (privKey interface{}, err error) {
	privKeyData, err := ioutil.ReadFile(privKeyPath)
	if err != nil {
		return nil, err
	}
	privKeyBlock, _ := pem.Decode(privKeyData)
	return x509.ParseECPrivateKey(privKeyBlock.Bytes)
}

// doRegister generates a CSR then asks Xbus for a registration.
func doRegister(renewKey, renewCert bool) int {
	xbusClient, err := GetClient(true)
	if err != nil {
		Logger.Errorf("%s", err)
		return 1
	}
	if err := xbusClient.Connect(); err != nil {
		Logger.Errorf("Cannot connect to server: %s", err)
		return 1
	}

	if err := xbusClient.CreateKey(renewKey); err != nil {
		if err != xbus.ErrAlreadyExists {
			Logger.Errorf("Cannot create a private key %s", err)
			return 1
		}
	}

	hostname := v.GetString("hostname")
	// TODO Read org and cn from the configuration
	if err := xbusClient.CreateCSR(hostname, "xbus", "xbusctl", renewCert); err != nil {
		if err != xbus.ErrAlreadyExists {
			Logger.Errorf("Cannot generate a CSR: %s", err)
			return 1
		}
	}

	// Send the registration request.
	regStatus, err := xbusClient.Register()
	if err != nil {
		Logger.Errorf("Cannot send the registration request: %s", err)
		return 1
	}

	Logger.Noticef("Successfully sent the registration request.")

	if renewCert && regStatus != api.Registration_PENDING {
		Logger.Errorf(
			"The server ignored the new CSR. This is most probably because " +
				"'account reset-cert' was not called. See the documentation " +
				"for more information: " +
				"https://docs.xbus.io/3.0/cert-maintenance.html#reset-the-cert",
		)
		return 1
	}

	// Say something depending on the registration status we have received in reply.
	switch regStatus {
	case api.Registration_PENDING:
		msg := fmt.Sprintf(`The registration is pending approval. To approve from Xbus, run:

	xbusctl account accept %q
`,
			xbusClient.Name(),
		)
		for _, actor := range xbusClient.ActorOptions() {
			msg += fmt.Sprintf("\txbusctl actor accept %q\n", actor.Name)
		}

		if renewCert {
			msg += fmt.Sprintf(
				"\nThen run the 'register' command\n")
		} else {
			msg += fmt.Sprintf(
				"\nThen run the exact same command you just runned (register)\n")
		}
		Logger.Noticef(msg)
	case api.Registration_ACCEPTED:
		Logger.Noticef(
			"The registration was already known by Xbus; it has been automatically accepted.",
		)
	case api.Registration_REJECTED:
		Logger.Noticef(
			"The registration was already known by Xbus; it has been automatically rejected.",
		)
	}

	return 0
}

// renewCertCmd is the "renew-cert" command.
var renewCertCmd = &cobra.Command{
	Use:   "renew-cert",
	Short: "Re-generate a CSR and ask Xbus for its registration.",
	Long: `Re-generate a Certificate Signing Request and optionnaly a private key,
then do to register`,
	Run: func(cmd *cobra.Command, args []string) {
		renewKey := cmd.Flag("newkey").Value.String() == "true"
		os.Exit(doRegister(renewKey, true))
	},
}

func init() {
	RootCmd.AddCommand(registerCmd)
	RootCmd.AddCommand(renewCertCmd)
	flags := renewCertCmd.PersistentFlags()
	flags.Bool("newkey", false, "Also generate a new private key")
}
