package fullstack

import (
	"context"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"

	xbus "bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/api/control"
	"bitbucket.org/orus-io/xbus/services/envelopestorage"
	"bitbucket.org/orus-io/xbus/testutils"
)

func restProcessCancel(t *testing.T) {
	ft := InitFullstack(t, true)
	defer ft.Shutdown()

	ft.Logger.ExpectsError(testutils.ExpectsContains(
		"[server] [director] Error calling agent jojo Process (attempt 1): Call canceled",
	))

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit := make(chan bool, 1)

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")

	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)

	defer cancel()

	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	xsub, err := consumer.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			t.Logf("Here i am. Now just waiting to be canceled")

			select {
			case <-apc.Context.Done():
				hit <- true
			case <-ctx.Done():
				hit <- false
			}
			return nil
		})
	assert.Nil(t, err)
	defer xsub.Unsubscribe()

	ft.InstallGraph(BuildSimpleGraph("evt1", "anyconsumer"))
	env := xbus.ForgeEnvelopeFromItems("evt1", []api.Item{[]byte("\"item1\"")})

	envStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer sub.Unsubscribe()

	status, err := emitter.SendEnvelope(nil, api.DefaultOutput, env)
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

	select {
	case <-ctx.Done():
		t.Errorf("Process state not updated")
		return
	case envState := <-envStateChan:
		assert.Equal(t, api.Process_RUNNING, envState.Status)
	}

	var pList []api.Process
	assert.Nil(t, ft.Client.ProcessQuery(
		ctx,
		api.ProcessFilter{Status: []api.Process_Status{api.Process_RUNNING}},
		func(l []api.Process) {
			pList = append(pList, l...)
		},
	))
	assert.Equal(t, 1, len(pList))
	assert.Equal(t, api.Process_RUNNING, pList[0].Status)

	time.Sleep(500 * time.Millisecond)

	err = ft.Client.ProcessControl(
		pList[0].ID, control.ProcessControlRequest_CANCEL)
	assert.Nil(t, err)

	select {
	case success := <-hit:
		if !success {
			t.Errorf("The processing was not canceled")
		}

	case <-ctx.Done():
		t.Errorf("Cancel failed")
	}
}

func TestProcessCancelOnStalledEnvelope(t *testing.T) {
	defer switchDurationVar(&envelopestorage.EnvelopeStallDelay, 1*time.Second)()
	ft := InitFullstack(t, true)
	defer ft.Shutdown()

	ft.Logger.ExpectsError(testutils.ExpectsContains(
		"[server] [director] Error calling agent jojo Process (attempt 1): Call canceled",
	))

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	canceled := make(chan bool, 1)

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")

	ctx, cancel := context.WithTimeout(context.Background(), 4*time.Second)

	defer cancel()

	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	xsub, err := consumer.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			t.Logf("Start reading the envelope so the stall is detected")

			receiver, err := apc.ReadEnvelope("default")
			if err != nil {
				return err
			}

			select {
			case <-apc.Context.Done():
				canceled <- true
			case <-time.After(2 * time.Second):
				canceled <- false
			}
			receiver.Cancel()

			ft.Logger.Noticef("consumer Process end")

			return nil
		})
	assert.Nil(t, err)
	defer xsub.Unsubscribe()

	ft.InstallGraph(BuildSimpleGraph("evt1", "anyconsumer"))
	env := xbus.ForgeEnvelopeFromItems("evt1", []api.Item{[]byte("\"item1\"")})
	env.Last = false
	env.Events[0].ItemCount = 2

	envStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer sub.Unsubscribe()

	status, err := emitter.SendEnvelope(nil, api.DefaultOutput, env)
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_RECEIVING, status)

	// The second item will never be sent. In about 15 seconds, the envelope
	// should be marked as STALLED, and the process canceled.

	select {
	case <-ctx.Done():
		t.Errorf("Process state not updated (expecting change to RUNNING)")
		return
	case envState := <-envStateChan:
		ft.Logger.Noticef("Process is RUNNING as expected")
		assert.Equal(t, api.Process_RUNNING, envState.Status)
	}

	select {
	case success := <-canceled:
		if !success {
			t.Errorf("The processing was not canceled")
		}
		ft.Logger.Noticef("The actor process was canceled as expected")

	case <-ctx.Done():
		t.Errorf("Cancel failed")
	}

	select {
	case <-ctx.Done():
		t.Errorf("Process state not updated (expecting change to ERROR)")
		return
	case envState := <-envStateChan:
		assert.Equal(t, api.Process_ERROR, envState.Status)
		ft.Logger.Noticef("Process is ERROR as expected")
	}
}
