package xbus

import (
	"errors"
	"sync"
	"time"

	api "bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
)

var (
	// ErrEventTypesNotKnown is returned by functions that work with the
	// event types if they were not yet received.
	ErrEventTypesNotKnown = errors.New("Event types are not known yet")

	// ErrWrongStatus is when addEnvelope is called with a current
	// status != api.EnvelopeAck_RECEIVING
	ErrWrongStatus = errors.New("Receiver is done receiving")

	// ErrTimeout is when a wait operation timed out
	ErrTimeout = errors.New("Timeout")

	// ErrReceptionError is returned by Complete() if current status is error
	ErrReceptionError = errors.New("Reception error")

	// ErrIncompleteEnvelope is when the source is closed and not all fragments
	// where received
	ErrIncompleteEnvelope = errors.New("Incomplete envelope")
)

// ItemQueue is a queue for items
// It is not thread safe, and would most probably benefit from a ring-buffer
// based implementation to avoid excessive reallocation when the items are
// consumed quickly
type ItemQueue []api.Item

// Add adds item(s) to the queue
func (q *ItemQueue) Add(items ...api.Item) {
	*q = append(*q, items...)
}

// Remove removes the first element and returns it
func (q *ItemQueue) Remove() api.Item {
	if len(*q) == 0 {
		panic("Remove called on empty queue")
	}
	item := (*q)[0]
	*q = (*q)[1:]
	return item
}

// RemoveN at most the N first elements and returns them
func (q *ItemQueue) RemoveN(n int) []api.Item {
	if len(*q) == 0 {
		panic("Remove called on empty queue")
	}
	if len(*q) < n {
		n = len(*q)
	}
	items := (*q)[:n]
	*q = (*q)[n:]
	return items
}

// RemoveS returns items with a total size not exceeding maxBytes
func (q *ItemQueue) RemoveS(maxBytes int) ([]api.Item, int) {
	var size, index int
	for i, item := range *q {
		nSize := size + len(item)
		if nSize <= maxBytes {
			size = nSize
			index = i
		} else {
			break
		}
	}
	return q.RemoveN(index + 1), size
}

// RemoveNS returns items with a total size not exceeding maxBytes,
// and a count not exceeding n
func (q *ItemQueue) RemoveNS(n, maxBytes int) ([]api.Item, int) {
	var size, index int
	for i, item := range *q {
		nSize := size + len(item)
		if i+1 == n {
			break
		}
		if nSize <= maxBytes {
			size = nSize
			index = i
		} else {
			break
		}
	}
	return q.RemoveN(index + 1), size
}

// EventValidatorMap is a map of EventValidator by EventID
type EventValidatorMap map[api.UUID]tools.EventValidator

// NewEnvelopeReceiver initialize a EnvelopeReceiver
func NewEnvelopeReceiver(input string, e api.Envelope) (*EnvelopeReceiver, error) {
	er := EnvelopeReceiver{
		ID:         e.ID,
		Input:      input,
		status:     api.EnvelopeAck_RECEIVING,
		validator:  tools.NewEnvelopeValidator(),
		itemQueues: make(map[api.UUID]ItemQueue),
		done:       make(chan struct{}),
	}
	er.cond = sync.NewCond(&er.mu)

	return &er, er.addEnvelope(e)
}

// NewEnvelopeReceiverSimple initialize a EnvelopeReceiver without an initial
// fragment
func NewEnvelopeReceiverSimple(id api.UUID) *EnvelopeReceiver {
	er := EnvelopeReceiver{
		ID:         id,
		Input:      "",
		status:     api.EnvelopeAck_RECEIVING,
		validator:  tools.NewEnvelopeValidator(),
		itemQueues: make(map[api.UUID]ItemQueue),
		done:       make(chan struct{}),
	}
	er.cond = sync.NewCond(&er.mu)

	return &er
}

// NewEnvelopeReceiverEmpty initialize a EnvelopeReceiver without an initial
// fragment, nor an envelopeID
func NewEnvelopeReceiverEmpty() *EnvelopeReceiver {
	return NewEnvelopeReceiverSimple(api.UUID{})
}

// EnvelopeReceiver handles an incoming envelope
type EnvelopeReceiver struct {
	ID        api.UUID
	Input     string
	status    api.EnvelopeAck_ReceptionStatus
	validator tools.EnvelopeValidator
	lastError error

	envelopeQueue []api.Envelope

	itemQueues   map[api.UUID]ItemQueue
	eventsByType map[string]api.UUID
	mu           sync.Mutex
	cond         *sync.Cond

	done chan struct{}

	OnCancel func()
}

// setDone mark the reception as over, and close the 'done' channel.
// Can be safely called several times, as long as the mu is locked by
// the caller
func (r *EnvelopeReceiver) setDone() {
	select {
	case _, isNotDone := <-r.done:
		if isNotDone {
			panic("done chan should never have any data")
		}
		// The chan is already closed, doing nothing
	default:
		// The channel is not closed, let's close it
		close(r.done)
	}
}

// Cancel should be called by the envelope consumer if for some reasons it
// decides not to consume the envelope anymore
func (r *EnvelopeReceiver) Cancel() {
	if r.OnCancel != nil {
		r.OnCancel()
	}
}

// Status returns the reception status
func (r *EnvelopeReceiver) Status() api.EnvelopeAck_ReceptionStatus {
	r.mu.Lock()
	defer r.mu.Unlock()
	return r.status
}

// LastError returns the reception status
func (r *EnvelopeReceiver) LastError() error {
	r.mu.Lock()
	defer r.mu.Unlock()
	return r.lastError
}

func (r *EnvelopeReceiver) mergeFragment(e api.Envelope) error {
	if r.ID.IsZero() {
		r.ID = e.ID
	}

	// validate
	if err := r.validator.Add(e); err != nil {
		r.setError(err)
		return err
	}

	if r.validator.EventTypesKnown && r.eventsByType == nil {
		r.eventsByType = make(map[string]api.UUID)
		for id, ev := range r.validator.Events {
			r.eventsByType[ev.Type] = id
		}
	}
	if r.validator.Complete {
		if err := r.validator.Check(); err != nil {
			r.setError(err)
			return err
		}
		r.status = api.EnvelopeAck_ACCEPTED
		r.setDone()
	}

	// put in event queues
	for _, event := range e.Events {
		q, _ := r.itemQueues[event.ID]
		q.Add(event.Items...)
		r.itemQueues[event.ID] = q
	}

	// signal all routines waiting for some change of state or
	// new items
	r.cond.Broadcast()

	return nil
}

func (r *EnvelopeReceiver) tryFlushEnvelopeQueue() error {
	for {
		var remains []api.Envelope
		for _, e := range r.envelopeQueue {
			if r.validator.IsNext(e) {
				if err := r.mergeFragment(e); err != nil {
					return err
				}
			} else {
				remains = append(remains, e)
			}
		}
		if len(remains) == len(r.envelopeQueue) {
			break
		}
		r.envelopeQueue = remains
	}
	return nil
}

func (r *EnvelopeReceiver) addEnvelope(e api.Envelope) error {
	r.mu.Lock()
	defer r.mu.Unlock()

	if r.status != api.EnvelopeAck_RECEIVING {
		return ErrWrongStatus
	}

	r.envelopeQueue = append(r.envelopeQueue, e)

	return r.tryFlushEnvelopeQueue()
}

// AddFragment add an envelope fragment
func (r *EnvelopeReceiver) AddFragment(e api.Envelope) error {
	return r.addEnvelope(e)
}

// Finalize signals the end of the fragments stream. If fragment(s) are missing,
// the receiver state will change to ERROR.
func (r *EnvelopeReceiver) Finalize() error {
	r.mu.Lock()
	defer r.mu.Unlock()

	if r.status == api.EnvelopeAck_ERROR {
		return ErrWrongStatus
	}

	if r.status == api.EnvelopeAck_RECEIVING {
		r.setError(ErrIncompleteEnvelope)
		return r.lastError
	}

	return nil
}

// SetError change the receiver state to ERROR with a message
func (r *EnvelopeReceiver) SetError(err error) {
	r.mu.Lock()
	defer r.mu.Unlock()

	r.setError(err)
}

// setError change the receiver state to ERROR with a message
func (r *EnvelopeReceiver) setError(err error) {
	r.status = api.EnvelopeAck_ERROR
	r.lastError = err
	r.cond.Broadcast()
	r.setDone()
}

func (r *EnvelopeReceiver) eventTypes() []string {
	if !r.validator.EventTypesKnown {
		return nil
	}
	var eventTypes []string
	for _, id := range r.validator.ExpectedEvents {
		ev := r.validator.Events[id]
		eventTypes = append(eventTypes, ev.Type)
	}
	return eventTypes
}

// EventTypes returns the types of event in the envelope, or nil if
// the types are not all known yet.
func (r *EnvelopeReceiver) EventTypes() []string {
	r.mu.Lock()
	defer r.mu.Unlock()

	return r.eventTypes()
}

// Events returns all information available about the events except their items.
// Returns nil if the event types are not known yet
func (r *EnvelopeReceiver) Events() []api.Event {
	r.mu.Lock()
	defer r.mu.Unlock()

	if !r.validator.EventTypesKnown {
		return nil
	}
	var events []api.Event

	for _, id := range r.validator.ExpectedEvents {
		ev := r.validator.Events[id]
		events = append(events, api.Event{
			ID:        ev.EventID,
			Type:      ev.Type,
			ItemCount: ev.ItemCount,
			Checksum:  ev.Checksum,
		})
	}
	return events
}

// WaitEventTypes returns the types of the event, or nil if the timeout
// expires and the types are still unknown
func (r *EnvelopeReceiver) WaitEventTypes(timeout time.Duration) []string {
	r.mu.Lock()
	defer r.mu.Unlock()

	var eventtypes = r.eventTypes()

	if eventtypes != nil || timeout == 0 {
		return eventtypes
	}

	deadline := time.Now().Add(timeout)

	// XXX I suspect a race-condition is possible if the timer is not precise
	// enough and hits the condition just before Now().Before(deadline) becomes
	// false
	// We need to investigate this a little more, maybe unittests will help
	timer := time.AfterFunc(timeout, func() {
		r.mu.Lock()
		r.cond.Broadcast()
		r.mu.Unlock()
	})
	defer timer.Stop()

	for eventtypes == nil && time.Now().Before(deadline) {
		r.cond.Wait()
		eventtypes = r.eventTypes()
	}

	return eventtypes
}

// waitNextItems returns at most 'max' items from the given eventType
// as soon a they are available. If not and the status is receiving, it waits
func (r *EnvelopeReceiver) waitNextItems(eventID api.UUID, max int) (items []api.Item) {
	r.mu.Lock()
	defer r.mu.Unlock()

	for r.status == api.EnvelopeAck_RECEIVING &&
		!r.validator.Events[eventID].Complete &&
		len(r.itemQueues[eventID]) == 0 {
		r.cond.Wait()
	}
	q, _ := r.itemQueues[eventID]
	if len(q) > 0 {
		items = q.RemoveN(max)
		r.itemQueues[eventID] = q
	}
	return
}

// NextItems retrieve at most 'max' items from the given eventType
func (r *EnvelopeReceiver) NextItems(eventType string, max int) (items []api.Item) {
	r.mu.Lock()
	defer r.mu.Unlock()

	eventID := r.eventsByType[eventType]

	q, _ := r.itemQueues[eventID]
	if len(q) > 0 {
		items = q.RemoveN(max)
		r.itemQueues[eventID] = q
	}

	return
}

// WaitCompletion waits until the full envelope was received.
// Returns false if the reception is not over after timeout
func (r *EnvelopeReceiver) WaitCompletion(timeout time.Duration) bool {
	select {
	case <-time.After(timeout):
		return false
	case <-r.done:
		return true
	}
}

// GetItemChan returns a channel on which items are sent by a goroutine
// as soon as they are received.
// The channel is closed as soon as the reception is stopped
// The function should be called once at most for each event type
func (r *EnvelopeReceiver) GetItemChan(eventType string) <-chan api.Item {
	c := make(chan api.Item)
	eventID := r.eventsByType[eventType]
	go func(r *EnvelopeReceiver, c chan<- api.Item) {
		for {
			items := r.waitNextItems(eventID, 100)
			if len(items) == 0 {
				// nothing more to expect, we close and quit
				close(c)
				return
			}
			for _, item := range items {
				c <- item
			}
		}
	}(r, c)
	return c
}

// Done is a chan that is closed when recetion is over
func (r *EnvelopeReceiver) Done() <-chan struct{} {
	return r.done
}

// Complete returns the complete envelope.
func (r *EnvelopeReceiver) Complete(timeout time.Duration) (*api.Envelope, error) {
	if !r.WaitCompletion(timeout) {
		return nil, ErrTimeout
	}
	r.mu.Lock()
	defer r.mu.Unlock()
	if r.status == api.EnvelopeAck_ERROR {
		if r.lastError != nil {
			return nil, r.lastError
		}
		return nil, ErrReceptionError
	}
	env := api.Envelope{
		ID:   r.validator.EnvelopeID,
		Last: true,
	}
	for _, id := range r.validator.ExpectedEvents {
		ev := r.validator.Events[id]
		env.EventIDs = append(env.EventIDs, id)
		env.Events = append(env.Events, api.Event{
			ID:        id,
			Type:      ev.Type,
			Index:     1,
			Items:     r.itemQueues[id],
			ItemCount: ev.ItemCount,
			Checksum:  ev.Checksum,
		})
	}
	return &env, nil
}
